/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Should you need to contact me, the author, you can do so by
 * e-mail - mail your message to Viktor T. Toth at <vttoth@vttoth.com>.
 *
 * ACKNOWLEDGEMENTS
 *
 * Many thanks to Paul Godfrey whose work provided the basis for this
 * implementation.
 */


#include <iostream>
#include <stdlib.h>
#include <string.h>
#include <gmp.h>

#include "mpcomplex.h"
#include "matrix.h"

using namespace std;

mpdouble Comb(int n, int k)
{
	mpdouble r;

	if (k < 0) return r;
	r = 1;
	for (int l = 1; l <= n-k; l++)
	{
		r *= n + 1 - l;
		r /= l;
	}
	return r;
}

void makeB(matrix<mpdouble> &B, int n)
{
	int r, c;
	for (c = 0; c < n; c++) B[0][c] = 1;
	for (r = 1; r < n; r++)
	{
		for (c = 0; c < n; c++)
		{
			B[r][c] = Comb(r+c-1, c-r);
			if ((r+c)&1) B[r][c] = -B[r][c];
		}
	}
}

void makeC(matrix<mpdouble> &C, int n)
{
	for (int i = 1; i < n; i++)
	{
		for (int j = 0; j <= i; j++)
		{
			C[i][j] = 0;
			for (int k = 0; k <= i; k++)
			{
				C[i][j] += Comb(2*i, 2*k) * Comb(k, k+j-i);
			}
			if ((i-j)&1) C[i][j] = -C[i][j];
		}
	}
	C[0][0] = 0.5;
}

void makeD(matrix<mpdouble> &D, int n)
{
	D[0][0] = 1;
	D[1][1] = -1;
	for (int i = 2; i < n; i++)
	{
		D[i][i] = D[i-1][i-1] * (2 * (2*i-1));
		D[i][i] /= (i - 1);
	}
}

void makeF(matrix<mpdouble> &F, mpdouble g, int n)
{
	for (int a = 0; a < n; a++)
	{
		int i;
		F[a][0] = 2;
		for (i = a + 1; i <= 2 * a; i++)
		{
			F[a][0] *= i;
			F[a][0] /= 4;
		}
		F[a][0] *= exp(mpdouble(a) + g + .5);
		F[a][0] /= pow(mpdouble(a) + g + .5, a);
		F[a][0] /= sqrt(mpdouble(a) + g + .5);
	}
}

int main(int argc, char *argv[])
{
	if (argc < 3)
	{
		cout << "Usage: " << argv[0];
		cout << " coeff terms [precision [ReArg [ImArg]]]" << endl;
		exit(0);
	}
	double g = atof(argv[1]);
	int n = atoi(argv[2]);
	if (argc > 3) mpdouble::set_default_prec(atoi(argv[3]));
	mpdouble zRe, zIm;
	bool loop = true;	// Do we use a user interface loop?
	if (argc > 4)
	{
		zRe = mpdouble(argv[4]);
		if (argc > 5)
		{
			zIm = mpdouble(argv[5]);
		}
		loop = false;
	}

	matrix<mpdouble> B(n, n);
	matrix<mpdouble> C(n, n);
	matrix<mpdouble> D(n, n);
	matrix<mpdouble> F(n, 1);
	matrix<mpdouble> P(n, 1);
	matrix<mpcomplex> Z(1, n);
	matrix<mpcomplex> R(1, 1);

	makeB(B, n);
	makeC(C, n);
	makeD(D, n);
	makeF(F, g, n);
	P = D*B*C*F;
	if (loop)
	{
		matrix<mpdouble> E(n, 1);
		cout << P;
		E = C * F;
		mpdouble err(0);
		for (int i = 0; i < n; i++)
		{
			if (i & 1) err -= E[i][0];
			else err += E[i][0];
		}
		err -= exp(mpdouble(g)) * sqrt(mpdouble::pi());
		err *= mpdouble::pi() / sqrt(mpdouble::e() * 2) / 2;
		cout << "|E|=" << abs(err) << endl;
	}

	do
	{
		if (loop)
		{
			cout << "Re Z=?";
			cout.flush();
			cin >> zRe;
			cout << "Im Z=?";
			cout.flush();
			cin >> zIm;
		}
		bool s = false;
		if (zRe < 0)
		{
			zRe = 1 - zRe;
			zIm = -zIm;
			s = true;
		}
		mpcomplex z(zRe - 1, zIm), r;
		Z[0][0] = mpdouble(1);
		for (int i = 1; i < n; i++)
		{
			Z[0][i] = 1/(z+i);
		}
		R = Z * P;
		r = z + g + .5;
		r = log(R[0][0]) + (z+.5) * log(r) - r;
		if (s)
		{
			r = log(mpdouble::pi()) - r - log(-sin(mpdouble::pi() * z));
		}
		if (loop)
			cout << "lnGma(z)=" << r << endl << "Gamma(z)=" << exp(r) << endl;
		else cout << exp(r).real() << endl;
	} while (loop);
}
